import typing
import collections.abc
from .communication import Command, Query, EventStream

CQProcessor = collections.abc.Callable[[Command | Query], EventStream]
CQProcessorContext = collections.abc.Mapping[typing.Any, typing.Any]


class CQRSFactory(typing.Protocol):
    async def get_processor(
        self,
        context: CQProcessorContext | None = None
    ) -> CQProcessor: ...



import typing
import abc
import collections.abc

T = typing.TypeVar("T")


class ClauseBit(typing.Generic[T]):
    def __init__(self, *data: typing.Self):
        self._data = data

    @property
    def value(self):
        return self._data

    def __add__(self, other: type[typing.Self]):
        if type(self) is ClauseBit:
            self._data = list(self._data)
            self._data.append(other)
            return self

        else:
            return ClauseBit(*(self, other))


class SliceBit(ClauseBit):
    def __init__(self, value: int):
        super().__init__()
        self._data = value

    @property
    def value(self):
        return self._data


class Limit(SliceBit):
    pass


class Offset(SliceBit):
    pass


class FilterBit(abc.ABC, ClauseBit[T]):
    def __init__(self, value: T):
        super().__init__()
        self._data = value

    @property
    def value(self):
        return self._data

    def __eq__(self, other: T):
        return self._data == other


class EQ(FilterBit[T]):
    pass


class NE(FilterBit[T]):
    pass


class LE(FilterBit[T]):
    pass


class GE(FilterBit[T]):
    pass


class LT(FilterBit[T]):
    pass


class GT(FilterBit[T]):
    pass


class IS(FilterBit[T]):
    pass


class LIKE(FilterBit[T]):
    def __init__(self, value: T, case_sensitive: bool = False):
        super().__init__(value)
        self._case_sensitive = case_sensitive

    def __eq__(self, other: typing.Self):
        return (
            self._data == other.value and
            self._case_sensitive == other.case_sensitive
        )

    @property
    def case_sensitive(self) -> bool:
        return self._case_sensitive


class SET(FilterBit[T]):
    def __init__(self, *value: T):
        super().__init__(set(value))


class RANGE(FilterBit[T]):
    def __init__(self, left: GE | GT | None, right: LE | LT | None):
        super().__init__((left, right))


class NOT(FilterBit):
    def __init__(self, value: FilterBit):
        super().__init__(value)
        self._value = value

    def __eq__(self, other):
        if other is None:
            return super().__eq__(other)
        else:
            return super().__eq__(other.value)


class OrderBit(abc.ABC, ClauseBit):
    def __init__(self, priority: int = 0):
        super().__init__()
        self._data = priority

    @property
    def value(self):
        return self._data


class ASC(OrderBit):
    pass


class DESC(OrderBit):
    pass


class Param(typing.Generic[T]):
    def __init__(self, name: str, value: ClauseBit[T]):
        self._value = value
        self._name = name

    @property
    def value(self) -> ClauseBit:
        return self._value

    @property
    def name(self) -> str:
        return self._name


NotationParser = typing.Callable[
    [str | collections.abc.Mapping[str, str]],
    collections.abc.Iterable[Param]
]



from . import query

from .base import (
    NoValueType
)

from .identity import (
    IdentifiersFactory
)

from .communication import (
    Message,
    Query,
    Context,
    Command,
    Event,
    IOEvent,
    BDEvent,
    StorageEvent,
    ResponseEvent,
    ReadEvent,
    WriteEvent,
    Error,
    Frame,
    EventStream
)

from .di import (
    DIResolverContract,
    DIContainerContract
)

from .cqrs import (
    CQProcessor,
    CQRSFactory
)



import collections.abc
import dataclasses
import re
import datetime
import typing
import types

import dateutil.parser
import functools

from .. import codex

FieldName = str
FieldType = type


@dataclasses.dataclass
class ParsingSchema:
    order_by: str = 'order_by'
    limit: str = 'limit'
    offset: str = 'offset'
    fieldset: str = 'fieldset'


def _cast_bool(value: str):
    value = value.strip().lower()
    if value == 'false':
        return False
    elif value == 'true':
        return True


default_casting_map = types.MappingProxyType({
    datetime.datetime: dateutil.parser.parse,
    datetime.date: datetime.date.fromisoformat,
    bool: _cast_bool
})

interval_types = (
    datetime.datetime,
    datetime.date,
    int,
    float
)


class Parser:
    def __init__(
        self,
        types_map: collections.abc.Mapping[FieldName, FieldType],
        casting_map: collections.abc.Mapping[
            FieldType,
            collections.abc.Callable[[str], typing.Any]
        ] = default_casting_map,
        parsing_schema: ParsingSchema = ParsingSchema()
    ):
        self._types_map = types_map
        self._casting_map = casting_map
        self._parsing_schema = parsing_schema

    def __call__(
        self,
        query: str | collections.abc.Mapping[str, str]
    ) -> collections.abc.Generator[codex.query.Param, None, None]:
        if isinstance(query, str):
            if '=' not in query:
                raise ValueError('You have to specify name for parameter value')
            query = (pairs.split('=') for pairs in query.split('&'))
        elif isinstance(query, collections.abc.Mapping):
            query = query.items()
        else:
            raise ValueError('Query mast be string or mapping')
        yield from self._parse(query)

    def _parse(self, query: collections.abc.Iterable[tuple[str, str]]):
        for pair in query:
            k, v = pair
            if v is None:
                continue
            if k == self._parsing_schema.order_by:
                yield from self._parse_order_param(v)
            elif k == self._parsing_schema.limit:
                yield codex.query.Param(
                    name=self._parsing_schema.limit,
                    value=codex.query.Limit(int(v))
                )
            elif k == self._parsing_schema.offset:
                yield codex.query.Param(
                    name=self._parsing_schema.offset,
                    value=codex.query.Offset(int(v))
                )
            else:
                yield self._parse_filter_param(k, v.strip())

    @staticmethod
    def _parse_order_param(
        names: FieldName
    ) -> collections.abc.Generator[codex.query.Param, None, None]:
        priority = 0
        for name in names.split(','):
            direction = codex.query.ASC
            if name.startswith('-'):
                direction = codex.query.DESC
                name = name[1:]
            yield codex.query.Param(
                name=name,
                value=direction(priority)
            )
            priority += 1

    def _parse_filter_param(self, name: FieldName, value: str):
        if name not in self._types_map:
            raise Exception(f'Type of parameter {name} must be defined in types map')

        for pattern, handler in self._pattern_handler_map.items():
            if mo := pattern.search(value):
                return handler(name, mo.group(1))

    def _interval(
        self,
        field_name: str,
        field_value: str,
        operations: tuple[
            type[codex.query.GT | codex.query.GE],
            type[codex.query.LT | codex.query.LE]
        ]
    ) -> codex.query.Param:
        if self._types_map[field_name] not in interval_types:
            raise TypeError(
                f'Interval cannot be calculated for type {self._types_map[field_name]} for field {field_name}')

        _data = field_value.split(',')
        if len(_data) != 2:
            raise ValueError(f'Range must contain strictly two members for field {field_name}')

        left = None
        right = None
        if _data[0]:
            left = operations[0](self._cast(_data[0], self._types_map[field_name]))
        if _data[1]:
            right = operations[1](self._cast(_data[1], self._types_map[field_name]))
        value = codex.query.RANGE(left, right)

        return codex.query.Param(name=field_name, value=value)

    def _multitude(
        self,
        field_name: str,
        field_value: str,
        inversion: bool = False
    ):
        field_value = codex.query.SET(
            *(
                self._cast(v, self._types_map[field_name])
                for v in field_value.split(',')
                if v
            )
        )

        if inversion:
            field_value = codex.query.NOT(field_value)

        return codex.query.Param(
            name=field_name,
            value=field_value
        )

    def _literal(
        self,
        field_name: str,
        field_value: str,
        operation: type[codex.query.FilterBit],
        inversion: bool = False
    ):
        field_value = operation(self._cast(field_value, self._types_map[field_name]))

        if inversion:
            field_value = codex.query.NOT(field_value)

        return codex.query.Param(
            name=field_name,
            value=field_value,
        )

    def _cast(self, value: str, type_: type):
        if cast := self._casting_map.get(type_):
            return cast(value)
        return type_(value)

    @functools.cached_property
    def _pattern_handler_map(self):
        return {
            re.compile('^(null)$'): lambda x, y: codex.query.IS(None),
            re.compile('^(!null)$'): lambda x, y: codex.query.NOT(codex.query.IS(None)),
            re.compile(r'^\(([\dTZ:\-,.]+)\)$'): functools.partial(
                self._interval,
                operations=(codex.query.GT, codex.query.LT)
            ),
            re.compile(r'^\[([\dTZ:\-,.]+)\)$'): functools.partial(
                self._interval,
                operations=(codex.query.GE, codex.query.LT)
            ),
            re.compile(r'^\(([\dTZ:\-,.]+)]$'): functools.partial(
                self._interval,
                operations=(codex.query.GT, codex.query.LE)
            ),
            re.compile(r'^\[([\dTZ:\-,.]+)]$'): functools.partial(
                self._interval,
                operations=(codex.query.GE, codex.query.LE)
            ),
            re.compile(r'^!{(.*)}$'): functools.partial(
                self._multitude,
                inversion=True
            ),
            re.compile(r'^{(.*)}$'): self._multitude,
            re.compile(r'^~{2}(.*)$'): functools.partial(
                self._literal,
                operation=codex.query.LIKE
            ),
            re.compile(r'^![~]{2}(.*)$'): functools.partial(
                self._literal,
                operation=codex.query.LIKE,
                inversion=True
            ),
            re.compile(r'^~(.*)$'): functools.partial(
                self._literal,
                operation=functools.partial(
                    codex.query.LIKE,
                    case_sensitive=True
                )
            ),
            re.compile(r'^!~(.*)$'): functools.partial(
                self._literal,
                operation=functools.partial(
                    codex.query.LIKE,
                    case_sensitive=True
                ),
                inversion=True
            ),
            re.compile(r'^!(.*)$'): functools.partial(
                self._literal,
                operation=codex.query.EQ,
                inversion=True
            ),
            re.compile(r'(.*)'): functools.partial(
                self._literal,
                operation=codex.query.EQ
            )
        }



import typing
import dataclasses
import collections.abc

Message = dataclasses.make_dataclass('Message', ())
Context = dataclasses.make_dataclass('Context', ())

Query = dataclasses.make_dataclass('Query', (), bases=(Message,))
Command = dataclasses.make_dataclass('Command', (), bases=(Message,))
Event = dataclasses.make_dataclass('Event', (), bases=(Message,))

Error = dataclasses.make_dataclass('Error', (), bases=(Event,))

IOEvent = dataclasses.make_dataclass('IOE', (), bases=(Event,))
BDEvent = dataclasses.make_dataclass('BDE', (), bases=(Event,))

StorageEvent = dataclasses.make_dataclass('StorageEvent', (), bases=(IOEvent,))
ResponseEvent = dataclasses.make_dataclass('ResponseEvent', (), bases=(IOEvent,))

ReadEvent = dataclasses.make_dataclass('ReadEvent', (), bases=(StorageEvent,))
WriteEvent = dataclasses.make_dataclass('WriteEvent', (), bases=(StorageEvent,))

P = typing.TypeVar('P', bound=Query | Command)
C = typing.TypeVar('C', bound=Context)


class Frame(typing.Generic[P, C]):
    payload: P
    context: C


EventStream = collections.abc.Iterable[Event]



import typing
import collections.abc

ContractType = typing.Any
ImplementationType = typing.Any

CallbackContextType = collections.abc.Mapping[str, typing.Any]
CallbackType = collections.abc.Callable[[ContractType, CallbackContextType], typing.NoReturn]


class DIResolverContract(typing.Protocol):
    async def resolve(
        self,
        contract: ContractType,
        execution_context: collections.abc.Mapping[str, typing.Any] | None = None,
        initial_context: collections.abc.Mapping[str, typing.Any] | None = None
    ): ...

    def add_context(
        self,
        contract: ContractType,
        data: collections.abc.Mapping
    ): ...

    async def shutdown(self, exc_type, exc_val): ...

    async def __aenter__(self): ...

    async def __aexit__(self, exc_type, exc_val, exc_tb): ...


class DIContainerContract(typing.Protocol):
    def register_dependency(
        self,
        implementation: ImplementationType,
        contract: ContractType | None = None,
        cache_scope: typing.Literal['container', 'resolver'] | None = None,
    ): ...

    def register_callback(
        self,
        contract: ContractType,
        callback: CallbackType,
        trigger: typing.Literal['shutdown'],
    ):  ...

    def get_resolver(self) -> DIResolverContract: ...

    async def shutdown(self, exc_type: typing.Any | None = None): ...

    async def __aenter__(self): ...

    async def __aexit__(self, exc_type, exc_val, exc_tb): ...
